/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.framework.config;

import java.io.*;

import java.net.URL;
import java.net.MalformedURLException;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.LogManager;

import sync4j.framework.config.*;
import sync4j.framework.tools.beans.BeanFactory;

//
// The following dependency is not good at all at this level. It is removed
// in SyncServer 4.1, but for now we keep it.
//
import sync4j.framework.engine.SyncStrategy;
import sync4j.framework.server.store.PersistentStore;
import sync4j.framework.security.Officer;


import org.apache.commons.lang.StringUtils;
import java.net.NetworkInterface;
import java.util.Enumeration;
import java.net.InetAddress;
import java.net.*;

/**
 * Incapsulates all the configuration information about the Sync4j Engine.
 * Configuration parameters are stored as properties in a Map structure. The
 * key is the name of the parameter and it is strucured in dotted separated
 * namespaces. The value of the parameter is the value of the key and can be of
 * any type. Accessor methods are provided to get the value of the parameter in
 * a particular type (such as String, int, double, boolean, ...).<br>
 * Access to getXXX() and setXXX() methods are synchronized.
 * <p>
 * This class follows the Singleton pattern, therefore it cannot be directly
 * instantiated; an internal instance is created just once and returned at any
 * getConfiguration() call.
 *
 * @author  Stefano Fornari @ funambol
 *
 * @version $Id: Configuration.java,v 1.8.4.12 2005/04/18 17:17:05 nichele Exp $
 */
public class Configuration
implements java.io.Serializable, ConfigurationConstants {

    // --------------------------------------------------------------- Constants

    protected static final String CONFIG_PATH  = File.separator + "config";
    protected static final String PROPERTY_SYSTEM_SYNC4J_HOME = "sync4j.home";

    // ------------------------------------------------------------ Private data

    private final Map internalMap = Collections.synchronizedMap(new HashMap());

    private static Configuration singleton;
    private transient ClassLoader classLoader = null;

    //
    // In order to have a logger completely independent by any configuration,
    // we use a standard anounymous logger. This will be used for error
    // notification only.
    //
    private static transient Logger log = Logger.getAnonymousLogger();

    // -------------------------------------------------------------- properties

    /** Getter for property classLoader.
     * @return Value of property classLoader.
     *
     */
    public ClassLoader getClassLoader() {
        return classLoader;
    }

    /** Setter for property classLoader.
     * @param classLoader New value of property classLoader.
     *
     */
    public void setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
    }

    // ------------------------------------------------------------ Constructors

    protected Configuration() {
    }

    public static synchronized Configuration getConfiguration() {
        if (singleton == null) {
            String sync4jProps = null;

            singleton = new Configuration();

            try {
                //
                // The following entry should be the URI, but in the case
                // it is not (it does not start with http://, https:// or file://,
                // it is considered simple file
                //
                URL configPath = fixURI(getConfigPath());

                singleton.setClassLoader(
                    new ConfigClassLoader(
                        new URL[] { configPath },
                        Configuration.class.getClassLoader()
                    )
                );

                sync4jProps =
                   configPath.toString() + File.separator + "Sync4j.properties";

                singleton.load(sync4jProps);

                configureLogging();

            } catch (ConfigurationException e) {
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe("Configuration error: " + e.getMessage());
                }
                log.throwing(Configuration.class.getName(), "getConfiguration", e);
            } catch (Exception e) {
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe("Error loading configuration from" + sync4jProps);
                }
                log.throwing(Configuration.class.getName(), "getConfiguration", e);
            }
        }

        return singleton;
    }

    // ---------------------------------------------------------- Public methods
    public String getStringValue(String name, String def)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        return (value == null) ? def : value;
    }

    public String getStringValue(String name)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        return value;
    }

    /**
     * Turns a character separated values property into an array of String
     * objects given the separator. For example:
     * <pre>
     *      one;two;trhee
     * </pre>
     * will result in
     * <pre
     *      {"one", "two", "trhee"}
     * </pre>
     *
     * @param name the name of the property
     * @param separator the separator character
     *
     * @return an array of Strings
     */
    public String[] getStringArray(String name, char separator)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        //
        // NOTE: String.split() is not used for performance reason. We do not
        //       need a regular expression here.
        //
        return StringUtils.split(value, String.valueOf(separator));
    }

    public int getIntValue(String name, int def)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        return (value == null) ? def : Integer.parseInt(value);
    }

    public int getIntValue(String name)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        return Integer.parseInt(value);
    }

    public boolean getBoolValue(String name, boolean def)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        return (value == null) ? def : Boolean.valueOf(value).booleanValue();
    }

    public boolean getBoolValue(String name)
    throws ConfigurationException {
        String value = (String)internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        return Boolean.valueOf(value).booleanValue();
    }

    public Object getValue(String name, Object def)
    throws ConfigurationException {
        Object value = internalMap.get(name);

        return (value == null) ? def : value;
    }

    public Object getValue(String name)
    throws ConfigurationException {
        Object value = internalMap.get(name);

        if (value == null) {
            throw new ConfigurationException("No " + name + " property defined");
        }

        return value;
    }

    /**
     * It is supposed that the value of the parameter is a class name. It returns
     * an instance of that class, created with <i>newInstance()</i>.
     *
     * @param name the name of the parameter
     *
     * @return an instance of the class specified by the parameter
     *
     * @throws ConfigurationException in case of errors
     */
    public Object getClassInstance(String name)
    throws ConfigurationException {
       String value = (String)internalMap.get(name);

       if (value == null) {
           throw new ConfigurationException("No " + name + " property defined");
       }

       try {
           return Class.forName(value).newInstance();
       } catch (Exception e) {
           throw new ConfigurationException( "Error in creating an instance of "
                                           + value
                                           , e                                );
       }
    }

    /**
     * It is supposed that the value of the parameter is a bean name. It returns
     * that bean, created with <i>BeanFactory.getBeanInstance()</i>.
     *
     * @param name the name of the parameter
     *
     * @return an instance of the class specified by the parameter
     *
     * @throws ConfigurationException in case of errors
     */
    public Object getBeanInstance(String name)
    throws ConfigurationException {
       String value = (String)internalMap.get(name);

       if (value == null) {
           throw new ConfigurationException("No " + name + " property defined");
       }

       try {
           return BeanFactory.getBeanInstance(classLoader, value);
       } catch (Exception e) {
           throw new ConfigurationException( "Error in creating an instance of "
                                           + value
                                           , e                                );
       }
    }

    /**
     * Create and return an array of bean as specified by a list of bean names.
     *
     * @param name the name of the paramenter
     * @param separator the separator character
     *
     */
    public Object[] getBeanArray(String name, char separator)
    throws ConfigurationException {
        String[] values = getStringArray(name, separator);

        Object[] ret = new Object[values.length];

        int i = 0;
        try {
            for (; i<ret.length; ++i) {
                ret[i] = BeanFactory.getBeanInstance(classLoader, values[i]);
            }
        } catch (Exception e) {
            throw new ConfigurationException ( "Error in creating the bean "
                                             + values[i]
                                             , e                           );
        }

        return ret;
    }

    /**
     * Loads and instantiate a bean by its config name. In this case the bean
     * is not looked up in the internal configuration map, but is created
     * directly by the means of the BeanFactory. This call is a shortcut for:
     * <code>
     *
     *  ClassLoader cl = Configuration.getConfiguration().getClassLoader();
     *  Object o = BeanFactory.getBeanInstance(cl, beanName);
     *
     * </code>
     *
     * @param beanName the complete beanName
     *
     * @return the bean instance
     *
     * @throws ConfigurationException in case of instantiation error
     */
    public Object getBeanInstanceByName(String beanName) {
        try {
             return BeanFactory.getBeanInstance(classLoader, beanName);
        } catch (Exception e) {
            String msg = "Error in creating an instance of " + beanName;

            log.severe(msg + ": " + e.getMessage());
            log.throwing(getClass().getName(), "init", e);

            throw new ConfigurationException(msg, e);
        }
    }

    public synchronized void load(InputStream is) throws IOException {
        Properties properties = new Properties();

        properties.load(is);

        internalMap.putAll(properties);

        setServerURI();
    }

    public synchronized void load(String uri)
    throws IOException {
        load(new URL(uri).openStream());
    }



    /**
     * Returns the Sync4j home as set as follows (the first value found is
     * taken):
     * <ul>
     *  <li>the system property sync4j.home</li>
     *  <li>the current directory</i>
     * </ul>
     *
     * @return the Sync4j home as set as discribed above.
     */
    public static String getSync4jHome() {
        return System.getProperty(PROPERTY_SYSTEM_SYNC4J_HOME, ".");
    }

    /**
     * Returns the config path calculated as getSync4jHome() + CONFIG_PATH
     *
     * @return the config path calculated as getSync4jHome() + CONFIG_PATH
     */
    public static String getConfigPath() {
        return getSync4jHome() + CONFIG_PATH;
    }

    /**
     * Configures the logging system accordingly to the
     * sync4j/server/logging/Logging.xml server bean.
     *
     * @throws IOException in case of errors setting up file handlers
     *
     */
    public static void configureLogging() throws IOException {
        //
        // NB: this will change in Sync4j 2.5!!!!!
        //

        //
        // First of all reinitialize logging
        //
        LogManager lm = LogManager.getLogManager();
        lm.reset();
        lm.readConfiguration();

        //
        // Now we can reconfigure Sync4j logging
        //
        Logger l;
        LoggerConfiguration conf;
        LoggingConfiguration lc = (LoggingConfiguration)
            singleton.getBeanInstanceByName(BEAN_LOGGING);

        Handler h;
        Iterator i = lc.getLoggers().iterator();
        while(i.hasNext()) {

            conf = (LoggerConfiguration)i.next();

            //
            // If this logger inherits from its parent, just skip logger
            // configuration.
            //
            if (conf.isInherit()) {
                continue;
            }

            l = Logger.getLogger(conf.getName());

            //
            // first remove all old handlers
            //
            Handler[] handlers = l.getHandlers();
            for (int j=0; ((handlers != null) && (j<handlers.length)); ++j) {
                handlers[j].flush();
                handlers[j].close();
                l.removeHandler(handlers[j]);
            }
            l.setUseParentHandlers(false);

            //
            // Now process the new handlers.
            // Note that at the handler stage we set level to ALL (use the
            // logger level to select the logging granularity.
            //
            if (conf.isConsoleOutput()) {
                h = new ConsoleHandler();
                h.setFormatter(new sync4j.framework.logging.SimpleFormatter());
                h.setLevel(Level.ALL);
                l.addHandler(h);
            }

            if (conf.isFileOutput()) {
                String pattern = conf.getPattern();

                if (new File(pattern).isAbsolute() == false) {
                    pattern = getSync4jHome() + File.separator + pattern;
                }

                if ((pattern == null) || (pattern.trim().length() == 0)) {
                    pattern = "logs/sync4j%t.log";
                }
                h = new FileHandler(
                    pattern        ,
                    conf.getLimit()*1048576,
                    conf.getCount(),
                    conf.isAppend()
                );
                h.setFormatter(new sync4j.framework.logging.SimpleFormatter());
                h.setLevel(Level.ALL);
                l.addHandler(h);
            }

            if (LoggerConfiguration.LEVEL_OFF.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.OFF);
            } else if (LoggerConfiguration.LEVEL_SEVERE.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.SEVERE);
            } else if (LoggerConfiguration.LEVEL_INFO.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.INFO);
            } else if (LoggerConfiguration.LEVEL_ALL.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.ALL);
            }

        }
    }

    /**
     * Creates and returns a new PersistentStore instance from the configured
     * <i>serverConfig.getEngineConfiguration().store</i> property
     *
     * @return the newly created instance
     */
    public PersistentStore getStore() {
        try {
            return (PersistentStore)getBeanInstance(ConfigurationConstants.CFG_PERSISTENT_STORE);
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the PersistentStore object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getStore", e);
        }

        return null;
    }

    /**
     * Creates and returns a new PersistentStore instance from the configured
     * <i>serverConfig.getEngineConfiguration().store</i> property
     *
     * @return the newly created instance
     */
    public Officer getOfficer() {
        try {
            return (Officer)getBeanInstance(ConfigurationConstants.CFG_SECURITY_OFFICER);
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the PersistentStore object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getStore", e);
        }

        return null;
    }

    /**
     * Creates and returns a new SyncStrategy instance from the configured
     * <i>serverConfig.getEngineConfiguration().strategy</i> property
     *
     * @return the newly created instance
     */
    public SyncStrategy getStrategy() {
        try {
            return (SyncStrategy)getBeanInstance(ConfigurationConstants.CFG_ENGINE_STRATEGY);
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the PersistentStore object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getStore", e);
        }

        return null;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();

        sb.append(getClass().getName()).append(" - ");
        sb.append(internalMap.toString()).append(" - ");
        sb.append("classLoader: " + classLoader);

        return sb.toString();
    }

    // --------------------------------------------------------- Private methods

    /**
     * Checks if the given string is an accepted URL (starting with http://,
     * https:// or file://). If yes, a new URL object representing the given
     * url is returned; otherwise, the given string is considered a file name
     * and a new URL is obtained calling File.toURL().
     *
     * @param s the string to check
     *
     * @return the corresponding URL if the string represents a URL or the
     *         fixed URL if the string is a pathname/filename
     *
     * @throws MalformedURLException
     */
    private static URL fixURI(final String s)
    throws MalformedURLException {

        try {
            return new URL(s);
        } catch (MalformedURLException e) {
            //
            // This is not a URL, let's consider it just a file
        }

        try {
            return new File(new File(s).getCanonicalPath()).toURL();

        } catch (IOException e) {
            throw new MalformedURLException("Unable to convert" + s + " to a URL");
        }
    }

    /**
     * The server URI is optional into Sync4j.properties: so if it's not given,
     * it should be set at default value http://<localhost>:8080/sync4j/sync
     * This is to avoid to change the configuration in order to access at the
     * server remotely. When server URI is specified, the given value must be
     * used instead of the default.
     */
    private void setServerURI() {
        try {
            String serverURI =
                (String)internalMap.get(ConfigurationConstants.CFG_SERVER_URI);
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Server URI set: " + serverURI);
            }

            if (serverURI.equals("")) {

                if (log.isLoggable(Level.WARNING)) {
                    log.warning(ConfigurationConstants.CFG_SERVER_URI + " property not provided; looking for local ip address.");
                }

                serverURI = "http://"
                          + getServerAddress()
                          + ":8080/sync4j/sync"
                          ;

                internalMap.put(ConfigurationConstants.CFG_SERVER_URI, serverURI);
            }
        } catch (Exception e) {
            String msg = "Error in setting default server URI";
            log.severe(msg + ": " + e.getMessage());
            log.throwing(Configuration.class.getName(), "setServerURI", e);
            throw new ConfigurationException(msg, e);
        }
    }

    /**
     * Get the server address checking all network interfaces. If there is an address
     * not loopback len 4 bytes return it, also return the host name. If a error
     * occurs getting the host name, returns <i>localhost</i>
     * @return String
     */
    private String getServerAddress() {
        NetworkInterface iface = null;
        InetAddress ia = null;
        String address = null;

        try {
            for (Enumeration ifaces = NetworkInterface.getNetworkInterfaces();
                                      ifaces.hasMoreElements(); ) {
                iface = (NetworkInterface)ifaces.nextElement();
                for (Enumeration ips = iface.getInetAddresses(); ips.hasMoreElements(); ) {
                    ia = (InetAddress)ips.nextElement();
                    if (!ia.isLoopbackAddress() && (ia.getAddress().length == 4)) {
                        address = ia.getHostAddress();

                        if (log.isLoggable(Level.WARNING)) {
                            log.warning("Local address <" + address + "> found on interface <" +
                                        iface + ">");
                        }
                        return address;
                    }

                }
            }
        } catch (SocketException ex) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe(ex.getMessage());
            }
            return null;
        }

        if (address == null) {
            if (log.isLoggable(Level.WARNING)) {
                log.warning("Non loopback address found, using local host name. " +
                            "If you want to access the server from an external network, please check your network " +
                            "configuration and/or provide a valid " +
                            ConfigurationConstants.CFG_SERVER_URI +
                            " property in Sync4j.properties. For local access, ignore this message.");
            }

            try {
                address = java.net.InetAddress.getLocalHost().getHostName();
            } catch (UnknownHostException ex) {
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe(ex.getMessage());
                }
                return null;
            }
        }
        return address;
    }

}
