/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server;

import java.io.Serializable;
import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.framework.logging.Sync4jLogger;

import sync4j.framework.core.SyncML;
import sync4j.framework.core.Sync4jException;

import org.apache.commons.lang.StringUtils;

/** 
 * This class is used for transformer the message in the canonical form.
 * @version $Id: SyncMLCanonizer.java,v 1.4.4.5 2005/01/29 13:53:43 luigiafassina Exp $
 */
public class SyncMLCanonizer implements Serializable {

    // ------------------------------------------------------------ Private data
    private static final Logger log = Sync4jLogger.getLogger("engine");
    
    // ------------------------------------------------------------ Constructors
    public SyncMLCanonizer() {
    }

    // ---------------------------------------------------------- Public methods
    
    /**
     * Canonizes outgoing messages.
     *
     * @param message the input XML message
     *
     * @return message the input XML message canonized
     **/
    public String canonizeOutput(String message) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Starting output canonization");
        }
        
        message = devInfHandler(message);
        
        return metInfNamespaceHandler(message);
    }
    
    /**
     * Canonizes incoming messages.
     *
     * @param message the input XML message
     *
     * @return message the input XML message canonized
     **/
    public String canonizeInput(String message) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Starting input canonization");
        }
        
        message = fixXMLVersion(message);
        
        return message;
    }
   
    /**
     * Replace into message the '&' with "&amp;"
     *
     * @param message the original message xml
     *
     * @return message the message updated
     */
    private String replaceEntity(String message) {
        return StringUtils.replace(message, "&", "&amp;");
    }
    
    /**
     * This is a temporary solution in order to obviate to a JiBX bug: it does 
     * not allow to declare the namespace to level of structure.
     * 
     * @param msg the server response
     *
     * @return the response with namespace correctly added into MetInf element
     */
    private String metInfNamespaceHandler(String msg) {
        int s = 0;
        int e = 0;
        
        StringBuffer response = new StringBuffer();
        while (( e = msg.indexOf("<Meta", s)) >= 0) {
            
            response = response.append(msg.substring(s, e));
            
            int a = msg.indexOf("</Meta>", e);
            String meta = msg.substring(e, a);
            
            meta = meta.replaceAll("<MetInf>" , "<MetInf xmlns='syncml:metinf'>");            
            meta = meta.replaceAll("<Type>"   , "<Type xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Format>" , "<Format xmlns='syncml:metinf'>");            
            meta = meta.replaceAll("<Mark>"   , "<Mark xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Size>"   , "<Size xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Anchor>" , "<Anchor xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Version>", "<Version xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<NextNonce>" , "<NextNonce xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<MaxMsgSize>", "<MaxMsgSize xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<MaxObjSize>", "<MaxObjSize xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<EMI>"    , "<EMI xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Mem>"    , "<Mem xmlns='syncml:metinf'>");
            
            s = a + 7;
            response.append(meta).append("</Meta>");
        }
        return response.append(msg.substring(s, msg.length())).toString();        
    }
    
    private String devInfHandler(String msg) {
        
        msg = StringUtils.replaceOnce(msg, "<UTC></UTC>", "<UTC/>");
        msg = StringUtils.replaceOnce(msg, "<SupportLargeObjs></SupportLargeObjs>", "<SupportLargeObjs/>");
        msg = StringUtils.replaceOnce(msg, "<SupportNumberOfChanges></SupportNumberOfChanges>", "<SupportNumberOfChanges/>");
        msg = StringUtils.replaceOnce(msg, "<MoreData></MoreData>", "<MoreData/>");
         
        return msg;
    }
    
    /**
     * Replaces the XML version from any value to "1.0", which is the version
     * supported by our parser. This should not cause any problem since SyncML
     * does not use any XML 1.1 specific feature.
     *
     * @param msg the message to process
     *
     * @return the processed message
     */
    private String fixXMLVersion(String msg) {
        
        if (!msg.startsWith("<?xml")) {
            return msg;
        }
        
        int e = msg.indexOf("?>");
        if (e < 0) {
            return msg;
        }
        
        String prolog = msg.substring(0, e+2);
        
        log.finest("prolog: " + prolog);
        
        prolog = prolog.replaceFirst("version(\\s)*=[\"'][^\"']+[\"']", "version=\"1.0\"");
        
        return prolog + msg.substring(e+2);
        
    }
}