/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.admin.ws.axis;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.axis.AxisFault;
import org.apache.axis.MessageContext;
import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.transport.http.HTTPConstants;
import org.apache.axis.utils.Messages;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.config.Configuration;
import sync4j.framework.config.ConfigurationConstants;
import sync4j.framework.server.Sync4jUser;
import sync4j.framework.server.store.PersistentStoreException;

import sync4j.server.admin.UserManager;

/**
 * This is an Axis authentication handler used to authenticate the user trying
 * to perform administration tasks on the Admin web service. Authentication is
 * currently performed with the Basic scheme.
 *
 * @author  Stefano Fornari @ Funambol
 * @version $Id: AdminAuthHandler.java,v 1.1.2.4 2005/01/26 17:17:15 nichele Exp $
 */
public class AdminAuthHandler
extends BasicHandler
implements ConfigurationConstants {

    // --------------------------------------------------------------- Constants

    public static final String LOG_NAME = "admin";

    public static final String ROLE_ADMIN = "sync_administrator";

    // ------------------------------------------------------------ Private data

    private Logger log = Sync4jLogger.getLogger(LOG_NAME);

    private UserManager userManager;

    // ------------------------------------------------------------ Constructors

    public AdminAuthHandler() {
        try {
            Configuration c = Configuration.getConfiguration();
            userManager = (UserManager)c.getBeanInstance(CFG_USER_MANAGER);
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error retrieving the user manager: " + e.getMessage());
            }
            log.throwing(AdminAuthHandler.class.getName(), "AdminAuthHandler", e);
        }
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Redefines the BasicHandler's invoke method.
     *
     * @param msgContext the Axis message context
     *
     * @throws AxisFault if authentication fails.
     */
    public void invoke(MessageContext msgContext) throws AxisFault
    {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Authenticating admin action.");
        }

        String username = msgContext.getUsername();
        if (username == null) {
            //
            // Hey! no auth here!
            // This will make axis return a 401
            //
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Credentials not provided.");
            }
            AxisFault f = new AxisFault();
            f.setFaultCodeAsString("Server.Unauthenticated");
            throw f;
        }

        String password = msgContext.getPassword();
        if (password == null) {
            password = "";
        }

        try {
            //
            // First of all: is the user authenticated?
            //
            Sync4jUser user = new Sync4jUser();
            user.setUsername(username);

            userManager.getUser(user);

            if (!password.equals(user.getPassword())) {
                //
                // Authentication failed!
                //
                notAuthorized();
            }

            //
            // Is the user a sync admin user?
            //
            userManager.getUserRoles(user);

            if (!user.hasRole(ROLE_ADMIN)) {
                //
                // The user is not an administrator
                //
                if (log.isLoggable(Level.INFO)) {
                    log.info( "Authorization denied: user "
                            + user
                            + " is not an administrator."
                            );
                }
                notAuthorized();
            }
        } catch (PersistentStoreException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error retrieving the user : " + e.getMessage());
            }
            log.throwing(AdminAuthHandler.class.getName(), "invoke", e);

            notAuthorized();
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * Throws the AxisFault representing "not authorized".
     *
     * @throws AxisFault
     */
    private void notAuthorized()
    throws AxisFault {
        AxisFault f = new AxisFault();
        f.setFaultCodeAsString("Server.Unauthorized");
        throw f;
    }
}
