/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.engine;

import java.io.*;
import java.util.logging.Logger;
import java.util.logging.Level;
import javax.naming.*;

import sync4j.framework.core.*;
import sync4j.framework.config.Configuration;
import sync4j.framework.config.ConfigurationConstants;
import sync4j.framework.engine.MessageSizeCalculator;
import sync4j.framework.engine.SyncEngineFactory;
import sync4j.framework.engine.pipeline.PipelineManager;
import sync4j.framework.engine.pipeline.MessageProcessingContext;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.protocol.ProtocolException;
import sync4j.framework.protocol.ProtocolUtil;
import sync4j.framework.server.SyncResponse;
import sync4j.framework.server.session.*;
import sync4j.framework.server.error.*;
import sync4j.framework.tools.beans.*;
import sync4j.framework.tools.WBXMLSizeCalculator;
import sync4j.framework.tools.WBXMLTools;
import sync4j.framework.tools.XMLSizeCalculator;

import sync4j.server.SyncMLCanonizer;
import sync4j.server.session.SyncSessionHandler;

import org.jibx.runtime.*;
import org.jibx.runtime.impl.*;

/**
 *  This class handles a synchronization request.
 *  <p>
 *  This server accepts synchronization requests addressed to the hostname
 *  indicated by the configuration property pointed by {CFG_SERVER_URI} (see
 *  Sync4j.properties).
 *  <p>
 *
 *  LOG NAME: sync4j.server
 *
 *  @author Luigia Fassina @ Funambol
 *
 *  @version $Id: SyncAdapter.java,v 1.10.4.11 2005/04/18 17:15:08 nichele Exp $
 *
 */
public class SyncAdapter implements ConfigurationConstants {

    // ------------------------------------------------------- Private constants
    protected static final String CFG_SYNCML_CANONIZER
    = "sync4j/server/SyncMLCanonizer.xml";

    protected static final String PARAM_SESSION_ID  = "sid";
    protected static final String PROPERTY_CONFIG_PATH = "sync4j.home";

    // ------------------------------------------------------------ Private data

    protected transient Logger log = null;

    protected SyncEngineFactory syncEngineFactory = null;

    protected SyncSessionHandler sessionHandler = null;

    protected Configuration config                = null;

    protected String sessionId                    = null;

    protected PipelineManager pipelineManager     = null;

    protected MessageProcessingContext mpc        = null;

    protected SyncML inputMessage                 = null;

    protected SyncMLCanonizer syncMLCanonizer     = null;

    protected String mimeType                     = null;
    protected String charSet                      = null;

    // ------------------------------------------------------------ Constructors

    /**
     * Creates a new SyncAdapter.
     *
     */
    public SyncAdapter() {
        super();

        log = Sync4jLogger.getLogger("server");

        loadConfiguration();

        sessionHandler = new SyncSessionHandler();

        pipelineManager = (PipelineManager)config.getBeanInstance(CFG_ENGINE_PIPELINE);

        try {
            syncMLCanonizer = (SyncMLCanonizer)BeanFactory.getBeanInstance(config.getClassLoader(), CFG_SYNCML_CANONIZER);
        } catch (Exception e){
            log.throwing("SyncAdapter", "constructor", e);
            new Sync4jException( "Error "
            + e.getClass().getName()
            + " creating the syncMLCanonizer: "
            + e.getMessage()
            ).printStackTrace();
        }

        mpc = new MessageProcessingContext();
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Sets sessionId
     *
     * @param sessionId the server session id
     */
    public void setSessionId(String sessionId) {
        this.sessionId = sessionId;
    }

    /**
     * Gets sessionId
     *
     * @return the server session id
     */
    public String getSessionId() {
       return sessionId;
    }

    /**
     * Finalizes the sync session
     */
    public void endSync() {
        int currentState = sessionHandler.getCurrentState();

        switch (currentState) {
            case SyncSessionHandler.STATE_END:
                sessionHandler.commit();
                break;

            case SyncSessionHandler.STATE_ERROR:
                sessionHandler.abort(StatusCode.PROCESSING_ERROR);
                break;

            default:
                sessionHandler.abort(StatusCode.SESSION_EXPIRED);
                break;
        }
        log = null;
    }

    /**
     * process the incoming sync message
     *
     * @param msg must be non-null
     * @param contentType may be null.
     *
     */
    public SyncResponse processMessage( final byte[] msg, final String contentType)
    throws ServerException {
        SyncResponse response = null;
        String inMessage = null;
        SyncML syncMLOut = null;

        if (msg == null) {
            String err = "msg is null!";
            log.severe(err);
            throw new ServerException(err);
        }

        if (contentType == null) {
            String err = "contentType is null!";
            log.severe(err);
            throw new ServerException(err);
        }

        mimeType    = contentType.split(";")[0];
        // Check if we have a charset.
        int pointer = contentType.indexOf(';');
        if (pointer > 0) {
            // Now strip of the 'charset=' by searching of the first '='.
            // Only one is allowed anyway.
            pointer = contentType.indexOf('=', pointer);
            if (pointer > 0) {
                String result = contentType.substring(pointer+1);
                // Just trim whitespace incase the client is not polite.
                charSet = result.trim();
            }
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("mimeType: " + mimeType);
            log.finest("charSet : " + charSet);
        }

        try {
            if (Constants.MIMETYPE_SYNCMLDS_WBXML.equals(mimeType)) {
                //
                // convert WBXML to XML, and then pass to processXMLMessage
                //
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Convert message from wbxml to xml");
                }
                // charSet may be null. In that case it is WBXML defined
                // and after that defaults to UTF-8.
                inMessage = WBXMLTools.wbxmlToXml(msg, charSet);

            } else if (Constants.MIMETYPE_SYNCMLDS_XML.equals(mimeType)) {
                // We check first the prolog for the encoding.
                charSet = getCharSetOfProlog(msg);
                if (charSet == null) { // If not in the prolog, use mime-type (charSet).
                    if (charSet == null) { // If still not defined default to UTF-8.
                        charSet = "UTF-8";
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Charset (default): " + charSet);
                        }
                    } else { // charset was defined in mime-type.
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Charset from mime-type: " + charSet);
                        }
                    }
                } else { // Charset was defined in the prolog.
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("Charset of prolog: " + charSet);
                    }
                }
                try {
                    inMessage = new String(msg, charSet);
                } catch (UnsupportedEncodingException e) {
                    // if we have a not known encoding try the default.
                    // FIXME, maybe better to bail out.
                    inMessage = new String(msg);
                }

            } else {
                //
                // its an unsupported MIME type
                //
                String err = "Unknown mime type:  " + mimeType;
                log.severe(err);
                throw new NotImplementedException(err);
            }

        } catch (Sync4jException e) {
            String err = "Error processing message: " + e.getMessage();
            log.severe(err);
            throw new ServerException(err);
        }

        inMessage = syncMLCanonizer.canonizeInput(inMessage);

        try {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Message for creating SyncML object\n" + inMessage);
            }

            IBindingFactory f = BindingDirectory.getFactory(SyncML.class);
            IUnmarshallingContext c = f.createUnmarshallingContext();

            inputMessage = (SyncML)c.unmarshalDocument(
                            new ByteArrayInputStream(inMessage.getBytes()), charSet);

        } catch(JiBXException e) {
            log.severe("Error unmarshalling message:" + e.getMessage());
            log.throwing(getClass().getName(), "processMessage", e);
            throw new ServerException(e);
        } catch(Exception e) {
            log.severe("Error processing input message: " + e.getMessage());
            log.throwing(getClass().getName(), "processMessage", e);
            throw new ServerException(e);
        }

        try {

            mpc.setProperty(mpc.PROPERTY_SESSIONID, sessionId);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Calling input pipeline");
            }
            pipelineManager.preProcessMessage(mpc, inputMessage);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Calling process message");
            }
            syncMLOut = processInputMessage();

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Calling output pipeline");
            }

            pipelineManager.postProcessMessage(mpc, syncMLOut);

            //
            // Set RespURI
            //
            syncMLOut.getSyncHdr().setRespURI(
                config.getStringValue(CFG_SERVER_URI) +
                '?' +
                PARAM_SESSION_ID +
                '=' +
                sessionId
            );

            byte[] out = null;
            if (Constants.MIMETYPE_SYNCMLDS_WBXML.equals(mimeType)) {
                //
                // The marshalling is doing directly into WBXMLTools
                // because that method is calling in the other code too
                //
                out = WBXMLTools.toWBXML(syncMLOut);

            } else if (Constants.MIMETYPE_SYNCMLDS_XML.equals(mimeType)) {
                ByteArrayOutputStream bout = new ByteArrayOutputStream();
                IBindingFactory f = BindingDirectory.getFactory(SyncML.class);
                IMarshallingContext c = f.createMarshallingContext();
                c.setIndent(0);
                c.marshalDocument(syncMLOut, charSet, null, bout);

                //
                // Temporary solution to handler namespace
                //
                out = syncMLCanonizer.canonizeOutput(bout.toString(charSet)).getBytes(charSet);
            }

            response = new Sync4jResponse(out, syncMLOut);

        } catch(JiBXException e) {
            log.severe("Error marshalling message:" + e.getMessage());
            log.throwing(getClass().getName(), "processMessage", e);
            throw new ServerException(e);
        } catch(Exception e) {
            log.severe("Error processing return message: " + e.getMessage());
            log.throwing(getClass().getName(), "processMessage", e);
            throw new ServerException(e);
        }

        return response;
    }

    /**
     * Returns the charset encoding if it was found in the prolog.
     * @param msg the message in which the prolog is searched.
     * @returns the charset if specified otherwise null.
     */
    private String getCharSetOfProlog(byte[] msg) {
        // Check if we have a start prolog, "<?xml".
        if ((msg[0] == '<') && (msg[1] == '?') && (msg[2] == 'x') && (msg[2] == 'm') && (msg[4] == 'l')) {
            // Now start looking for the end of the prolog.
            int length = msg.length - 1; // substract one for the double byte check.
            for (int i = 5; i < length; i++) {
                 if ((msg[i] == '?') && (msg[i+1] == '>')) {
                     // We found the end of the prolog.
                     String prolog = new String(msg, 0, i+1);
                     if (log.isLoggable(Level.FINEST)) {
                         log.finest("Message prolog: " + prolog);
                     }
                     // Search for encoding keyword.
                     int p = prolog.indexOf("encoding");
                     // First double quote after the encoding.
                     int q1 = prolog.indexOf('"', p);
                     // Second double quote after the encoding.
                     int q2 = prolog.indexOf('"', q1);
                     return prolog.substring(q1, q2);
                 }
            }
        }
        return null; // Was not defined
    }

    /**
     * Used to process a status information as needed by the client object (i.e.
     * errors or success).
     *
     * @param statusCode the status code
     * @param statusMessage additional descriptive message
     *
     * @see sync4j.framework.core.StatusCode for valid status codes.
     */
    public SyncResponse processStatusCode(int statusCode, String info){
        if (statusCode != StatusCode.OK) {
            sessionHandler.abort(statusCode);
        }
        return null;
    }

    // --------------------------------------------------------- private methods

    /**
     * Processes the input SyncML message after convertion to objects. See class
     * description for more information.
     *
     * @return the response message
     *
     * @throws ServerException in case of errors
     */
    private SyncML processInputMessage() throws ServerException {
        try {
            try {
                sessionHandler.setSizeCalculator(getSizeCalculator());
                return sessionHandler.processMessage(inputMessage, mpc);
            } catch (InvalidCredentialsException e) {
                return sessionHandler.processError(inputMessage, e);
            } catch (ServerException e) {
                return sessionHandler.processError(inputMessage, e);
            } catch (ProtocolException e) {
                return sessionHandler.processError(inputMessage,
                                       new BadRequestException(e.getMessage()));
            }

        } catch (Sync4jException e1) {
            //
            // This can be due only to processError
            //
            throw new ServerException(e1);
        }
    }

    /**
     * Loads the configuration for the server. The URI from which the configuration
     * must be set is <i>PROPERTY_CONFIG_PATH</i>
     *
     * @throws ConfigurationException in case of errors.
     */
    private void loadConfiguration() {

        config = Configuration.getConfiguration();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Configuration: " + config);
        }
    }

    /**
     * Returns the MessageSizeCalculator to be used accordingly to the message
     * content type (retrieved by the processing context)
     *
     * @return
     */
    private MessageSizeCalculator getSizeCalculator() {
        MessageSizeCalculator calculator = null;

        if ((mimeType != null) && Constants.MIMETYPE_SYNCMLDS_WBXML.equals(mimeType)) {
            calculator = new WBXMLSizeCalculator();
        } else {
            calculator = new XMLSizeCalculator();
        }

        return calculator;
    }

    // ------------------------------------------------------------ SyncResponse

    private static class Sync4jResponse implements SyncResponse {
        private byte[] msg;
        private SyncML syncML;

        /**
         * Create a new Sync4jResponse.
         *
         * @param msg the marshalled response
         * @param syncML the SyncML object
         * @param resultMimeType the mime type
         *
         */
        private Sync4jResponse(final byte[]  msg   ,
                               final SyncML  syncML) {
            this.msg            = msg;
            this.syncML         = syncML;
        }

        public byte[] getMessage() {
            return this.msg;
        }

        public String getMessageString() {
            return Util.toXML(syncML);
        }

        /**
         * Is this message the last message allowed for the current session?
         *
         * @return true if yes, false otherwise
         */
        public boolean isCompleted() {
            return (syncML.isLastMessage() &&
                    ProtocolUtil.noMoreResponse(syncML) &&
                    (ProtocolUtil.getStatusChal(syncML) == null));
        }
    }
}
